/**
  *
  * Klasse GraphInit liest Graphendaten aus TXT-Datei
  * die TXT-Datei ist von GraphenDemo oder manuell erzeugt
  * und enthlt Zeilen mit komma-sep. Knoten-Daten
  * im Format KnotenId,Name,x,y (z. B. 1,S1,3,1)
  * gefolgt von drei Zeilen mit Info zur Gerichtetheit
  * ---------------------------
  * Symmetrie = 1
  * ---------------------------
  * gefolgt von Zeilen mit komma-sep. Kanten-Daten
  * im Format KnotenId1,KnotenId2,Gewicht (z. B. 1,3,12)
  *
  * Version 1.0 enthlt noch nicht die implementierte Methode setGraph()
  * Version 2.0 enthlt die implementierte Methode setGraph()
  * Version 2.1 enthlt Methoden breitenDurchlauf und tiefenDurchlauf()
  * Version 3.1 bildet GraphInit-Attribut mit XVertex-Objekten()
  *
  * @version 2.1 vom 04.07.2017
  * @version 3.1 vom 10.07.2017
  * @author Michael Ziemke, Koeln & Milan Paessler (Q1-IF-LK1-ZMK)
  */
  
import java.io.BufferedReader;
//import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
//import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;

public class GraphInit {
  
  // Anfang Attribute
  private final boolean MIT_XVERTEX = true;
  private String name;
  private Graph meinGraph;
  private int[][] meineMatrix;
  private String[] meineKnoten;
  private String[] meineKnotenKoord;
  private boolean istUngerichtet = true;
  private String[] iniText;
  // Ende Attribute
  
  // Anfang Methoden
  /**
   * Konstruktor mit bergabe des IniTextes und Graphname
   * @param pIniText (Feld von Strings) Zeilen mit Knoten- und Kantendaten
   * @param pName (String) Name des neuen Graphen
   */
  public GraphInit (String[] pIniText, String pName) {
    this.name = pName;
    if (this.erzeugeIniText(pIniText))
      this.init();
  }
    
  /**
   * Konstruktor mit bergabe von Ordnername und IniTxt-Dateiname
   * @param pFolder (String) relativer Ordnername (z.B. ".\\initext")
   * @param pFile (String) Name der IniText-Datei (z.B. "graph1.txt")
   */
  public GraphInit(String pFolder, String pFile) {
    this.name = pFile;
    if (this.leseIniTextAusDatei(pFolder, pFile))
      this.init();
  } 
  
  /**
   * ruft bei Existenz der IniText-Daten die Methoden zur Erzeugung
   * der Datenreprsentanzen zu Knoten Kanten und Gerichtetheit und
   * zur Erzeugung des Graph-Objektes (NRW-ZA ab 2018) auf
   */
  private void init ()
  {
    this.erzeugeKnotenAusIni();
    this.erzeugeGerichtetAusIni();
    this.erzeugeMatrixAusIni();
    this.erzeugeGraph();  // Graph-Objekt mit Vertex- bzw. XVertex-Objekten
  }
    
  /**
   * liefert den Namen des Graphen
   * @return (String) Name des Graphen
   */
  public String gibName() {
    return this.name;
  }

  /**
   * liefert den IniText des Graphen
   * @return (Feld von Strings) Graphdaten: Knoten, Gerichtetheit, Kanten
   */
  public String[] gibIniText() {
    return this.iniText;
  }

  /**
   * bernimmt bergebene Graph-Daten als Feld von Strings
   * Zeilen-Formate zu einem Knoten: ID-Nummer, ID-Text, x-Koord, y-Koord (z.B. 1,S1,3,0)
   * Zeilen-Format zu Gerichtetheit: Zeile mit Text: Symmetrie = 1 oder Symmetrie = 0
   * Zeilen-Formate zu einer Kante: KnotenID-Nummer1, KnotenID-Nummer2, Gewicht (z.B. 1,3,12)
   * Diese Formate entsprechen denen der Graphendaten des Programms GraphenDemo
   * von Daniel Garmann, Gymnasium Odenthal; nur noch hier erhltlich:
   * http://www.ziemke-koeln.de/download/lk13if/graphentheorie/graphendemo.zip
   * @param pIniText (Feld von Strings) die Graphen-Daten
   * @return (boolean) true, falls Daten bernommen wurden
   */
  public boolean erzeugeIniText(String[] pIniText) {
    this.iniText = pIniText;
    return this.iniText.length > 0;
  }
  
  /**
   * bernimmt bergebene Graph-Daten als Feld von Strings
   * Zeilen-Formate zu einem Knoten: ID-Nummer, ID-Text, x-Koord, y-Koord (z.B. 1,S1,3,0)
   * Zeilen-Format zu Gerichtetheit: Zeile mit Text: Symmetrie = 1 oder Symmetrie = 0
   * Zeilen-Formate zu einer Kante: KnotenID-Nummer1, KnotenID-Nummer2, Gewicht (z.B. 1,3,12)
   * Die Koordinaten sind fr eine geplante graphische Reprsentanz ntig.
   * Diese Formate entsprechen denen der Graphendaten des Programms GraphenDemo
   * von Daniel Garmann, Gymnasium Odenthal; nur noch hier erhltlich:
   * http://www.ziemke-koeln.de/download/lk13if/graphentheorie/graphendemo.zip
   * @param pIniText (Feld von Strings) die Graphen-Daten
   * @return (boolean) true, falls Daten bernommen wurden
   */
  public boolean leseIniTextAusDatei(String pFolder, String pFile) {
    // mit Einbindung der Klasse Textdatei:
    // Textdatei lReader = new Textdatei(pFolder, pFile);
    // this.iniText = lReader.getLines();
    
    // ohne Klasse Textdatei:
    // WICHTIG: Java-Importe (siehe oben vor Klassenkopf)
    int lAnz = 0;
    File lFile = null;
    String lZeile = null;
    BufferedReader myReader;
    ArrayList<String> lArrayList = new ArrayList<String>();

    try 
    {
      lFile = new File(pFolder + "\\" + pFile);
      // Datei zum Lesen ffnen:
      myReader = new BufferedReader(new FileReader(lFile));
      // Textdatei zeilenweise einlesen, Elementanzahl ermitteln:
      while ((lZeile = myReader.readLine()) != null) 
      {
        lArrayList.add(lZeile);
      }
      // Feld-Attribut neu dimensionieren:
      lAnz = lArrayList.size();
      this.iniText = new String[lAnz];
      // Elemente in StringFeld-Attribut schreiben aus der ArrayList:
      for (int i=0 ; i<lAnz ; i++) 
      {
        this.iniText[i] = lArrayList.get(i);
      } // end of for
      myReader.close();
      
    } catch(Exception e) {
      System.out.println("##### Fehler: Datei fehlt: " + pFolder + "\\" + pFile);
      System.out.println("##### oder Fehler beim Auslesen der Datei ");
      System.out.println("##### Fehler-Details " + e.toString());
    } finally {
      // System.out.println("### DEBUG: gefunden=" + lFile.exists());
    } // end of try
    return this.iniText.length>0;
  }
  
  /**
   * liest die Art der Gerichtetheit aus den IniText-Daten und legt sie fest
   */
  private void erzeugeGerichtetAusIni() {
    int lZeileNr = 0;
    this.istUngerichtet = true;
    while (lZeileNr < this.iniText.length 
      && !this.iniText[lZeileNr].startsWith("Symmetrie"))
    {
      lZeileNr++;
    }
    if (this.iniText[lZeileNr].startsWith("Symmetrie")) {
      this.istUngerichtet = this.iniText[lZeileNr].equals("Symmetrie = 1");
    } // end of if
  }  
  
  /**
   * liest die Daten aller Knoten aus den IniText-Daten
   */
  private void erzeugeKnotenAusIni() {
    // aus Ini-Text:
    int lAnzahl = 0;
    // Knotenanzahl ermitteln:
    while (this.iniText.length>0 && lAnzahl < this.iniText.length 
      && this.iniText[lAnzahl].charAt(0) != '-' 
      && !this.iniText[lAnzahl].startsWith("Symmetrie"))
    {
      lAnzahl++;
    }
    // meineKnoten initialisieren; erster hat Index 1:
    meineKnoten = new String[lAnzahl+1];
    this.meineKnotenKoord = new String[lAnzahl+1];
    // Nummer und Bezeichnung jedes Knotens speichern:
    for (int i=0; i<lAnzahl; i++) {
      String[] lTemp = this.iniText[i].split(",");
      int lPos = Integer.parseInt(lTemp[0]);
      this.meineKnoten[lPos] = lTemp[1];
      if (lTemp.length > 3) {
        this.meineKnotenKoord[lPos] = lTemp[2] + "," + lTemp[3];
      }
    } // end of for
  }
  
  /**
   * liefert die Koordinaten-Daten eines oder aller Knoten
   * @param (int) Nummer des Knotens
   * @return (Feld von String) x- und y-Koordinate z. B. 1,3 ; fr grafische Reprsentanz
   */
  public String[] gibKnotenKoord(int pNummer) {
    String[] lRet;
    if (pNummer > 0 && pNummer < this.meineKnoten.length) {
      // Koordinaten eines Knotens:
      lRet = new String[1];
      lRet[0] = this.meineKnotenKoord[pNummer];
    } else {
      // Koordinaten aller Knoten:
      lRet = new String[this.meineKnotenKoord.length];
      for (int i=1 ; i<this.meineKnotenKoord.length ; i++) {
        lRet[i] = this.meineKnotenKoord[i]; 
      } // end of for
    } // end of if-else
    return lRet;
  }  

  /**
   * liefert die Koordinaten-Daten eines oder aller Knoten
   * @param (int) Nummer des Knotens; 0 fr alle
   * @return (Feld von String) im Format Name (x,y) z. B. S1 (1,2) ; fr Datenkontrolle
   */
  public String[] gibKnotenKoordAlsStrings(int pNummer) {
    String[] lRet;
    if (pNummer > 0 && pNummer < this.meineKnoten.length) {
      // Koordinaten eines Knotens:
      lRet = new String[1];
      lRet[0] = this.meineKnoten[pNummer] + " (" +this.meineKnotenKoord[pNummer] + ")";
    } else {
      // Koordinaten aller Knoten:
      lRet = new String[this.meineKnotenKoord.length];
      for (int i=1 ; i<this.meineKnotenKoord.length ; i++) {
        lRet[i] = this.meineKnoten[i] + " (" + this.meineKnotenKoord[i] + ")"; 
      } // end of for
    } // end of if-else
    return lRet;
  }  

  /**
   * liest die Daten aller Kanten aus den IniText-Daten
   * und erzeugt daraus als Attribut eine Adjazenz-Matrix-Reprsentanz
   */
  private void erzeugeMatrixAusIni() {
    // erste Zeile mit Kantenangabe ermitteln:
    int k1, k2, g;  // KnotenNr1, KnotenNr2, Gewicht
    int lZeileNr = this.meineKnoten.length + 2; // erste Zeile mit Kantenangaben
    this.meineMatrix = new int [this.meineKnoten.length][this.meineKnoten.length];
    // Initialisieren mit Wert -1:
    for (int i=0 ; i<this.meineKnoten.length ; i++) 
      for (int j=0 ; j<this.meineKnoten.length ; j++) 
        this.meineMatrix[i][j] = -1;
    // Auslesen der Gewichte:    
    while (this.iniText.length > lZeileNr)
    {
      String[] lTemp = this.iniText[lZeileNr].split(",");
      k1 = Integer.parseInt(lTemp[0]);
      k2 = Integer.parseInt(lTemp[1]);
      g = -1;
      if (lTemp[2] != "") {
        g = Integer.parseInt(lTemp[2]);
      } // end of if
      this.meineMatrix[k1][k2] = g;
      if (this.istUngerichtet) {
        this.meineMatrix[k2][k1] = g;
      } // end of if
      lZeileNr++;
    }
    
  }
  
  /**
   * liefert die Anzahl der Knoten des Graphen
   * @return (int) Anzahl der Knoten
   */
  public int gibDimension() {
    return this.meineKnoten.length;
  }
  
  /**
   * liefert die amen aller Knoten als kommaseparierten String
   * @return (String) komma-separierte Knotennamen
   */
  public String gibKnotenFolge()
  {
    String lRet = "";
    for (int i=1 ; i<this.meineKnoten.length ; i++) 
      lRet += "," + this.meineKnoten[i];
    return lRet;
  }
        
  /**
   * liefert die Adjazenz-Matrix-Reprsentanz des Graphen
   * @return (zweidimensionales Feld aus int) Matrix mit Kanten-Gewichten
   */
  public int[][] gibMatrix() {
    return this.meineMatrix;
  }  
  
  /**
   * erzeugt als Attribut ein Graph-Objekt mit Knoten und Kanten
   * gem der NRW-Java-Abiturklassen Graph, Vertex und Edge (ab 2018)
   */
  private void erzeugeGraph() {
    this.meinGraph = new Graph();
    int lKnotenAnz = this.meineKnoten.length;
    if (MIT_XVERTEX) {
      for (int i=1 ; i<lKnotenAnz ; i++) 
      {
        String[] tmp = this.meineKnotenKoord[i].split(",");  // String-Feld (x,y)
        this.meinGraph.addVertex( new XVertex(
          this.meineKnoten[i], 
          Integer.parseInt(tmp[0].trim()), 
          Integer.parseInt(tmp[1].trim())
          ));
      }  
    } else {
      for (int i=1 ; i<lKnotenAnz ; i++) 
        this.meinGraph.addVertex(new Vertex(this.meineKnoten[i]));
    } // end of if-else
    for (int i=1 ; i<lKnotenAnz ; i++) 
      for (int j=1 ; j<lKnotenAnz ; j++) 
        if (this.meineMatrix[i][j] != -1) 
          this.meinGraph.addEdge(new Edge(
            this.meinGraph.getVertex(this.meineKnoten[i]), 
            this.meinGraph.getVertex(this.meineKnoten[j]), 
            this.meineMatrix[i][j]
          ));
  }
  
  /**
   * erzeugt als Attribut ein Graph-Objekt mit Knoten und Kanten
   * gem der NRW-Java-Abiturklassen Graph, Vertex und Edge (ab 2018)
   * aber mit XVertex- statt Vertex-Objekten fr GUI-Darstellung
   */
  /* // OBSOLET 
  private void erzeugeXGraph() {
    this.meinGraph = new Graph();
    int lKnotenAnz = this.meineKnoten.length;
    for (int i=1 ; i<lKnotenAnz ; i++) 
    {
      String[] tmp = this.meineKnotenKoord[i].split(",");  // String-Feld (x,y)
      this.meinGraph.addVertex( new XVertex(
        this.meineKnoten[i], 
        Integer.parseInt(tmp[0].trim()), 
        Integer.parseInt(tmp[1].trim())
        ));
    }  
    for (int i=1 ; i<lKnotenAnz ; i++) 
      for (int j=1 ; j<lKnotenAnz ; j++) 
        if (this.meineMatrix[i][j] != -1) 
          this.meinGraph.addEdge(new Edge(
            this.meinGraph.getVertex(this.meineKnoten[i]), 
            this.meinGraph.getVertex(this.meineKnoten[j]), 
            this.meineMatrix[i][j]
          ));
  }
  */
  
  /**
   * erzeugt den Code, mit dem ein Objekt der Klasse Graph erstellt werden kann
   * gem der NRW-Java-Abiturklassen Graph, Vertex und Edge (ab 2018)
   * @author Milan Pssler (Q1-IF-LK1-ZMK)
   */
  public String erzeugeGraphCode(String pName) {
    String lRet = "Graph " + pName + " = new Graph();\n\n";
  
    this.meinGraph = new Graph();
    int lKnotenAnz = this.meineKnoten.length;
    
    for (int i=1 ; i<lKnotenAnz ; i++)
    { 
      lRet += pName;
      lRet += ".addVertex(new Vertex(\"";
      lRet += this.meineKnoten[i];
      lRet += "\"));\n";
    }     
                        
    lRet += "\n";
    
    for (int i=1 ; i<lKnotenAnz ; i++) 
      for (int j=1 ; j<lKnotenAnz ; j++)
        if (this.meineMatrix[i][j] != -1)
        {         
          lRet += pName;
          lRet += ".addEdge(new Edge(";
          lRet += pName;
          lRet += ".getVertex(\"";
          lRet += this.meineKnoten[i];
          lRet += "\"), ";            
          lRet += pName;
          lRet += ".getVertex(\"";
          lRet += this.meineKnoten[j];
          lRet += "\"), ";
          lRet += this.meineMatrix[i][j];
          lRet += "));\n";
        }
      
    return lRet;
  }
  
  /**
   * liefert das aus den IniText-Daten erzeugte Graph-Objekt (NRW-Java ab 2018)
   * z. B. zur weiteren Verarbeitung (Dijkstra, Kruskal, Prim, Euler)
   * @return (Graph) das erzeugte Graph-Objekt (gem NRW-Java ab 2018)
   */
  public Graph gibGraph()
  {
    return this.meinGraph;
  }
  /**
   * liefert die Daten des Graph-Objekts (Knoten und Kanten des Graphen)
   * z. B. fr eine Kontroll-Anzeige
   * @return (Feld aus Strings) Daten der Knoten und Kanten
   * @author Michael Ziemke, Koeln
   */
  public String[] gibGraphAlsStrings()
  {
    String[] lRet = {"FEHLER: Keine Graph-Daten gefunden!"};
    List<Vertex> lKnoten = new List<Vertex>();
    List<Edge> lKanten = new List<Edge>();
    if (this.meinGraph != null) {
      lKnoten = this.meinGraph.getVertices();
      lKanten = this.meinGraph.getEdges();
    }  
    lKnoten.toFirst();
    int lNr = 0;
    int lPos = 0;
    while (lKnoten.hasAccess()) {
      lPos++;
      lKnoten.next(); 
    }
    lKanten.toFirst();
    while (lKanten.hasAccess()) {
      lPos++;
      lKanten.next(); 
    }
    // Rckgabe-String-Feld ggf. neu dimensionieren:
    if (this.meinGraph != null)
      lRet = new String[lPos];
    
    // Knotendaten des Graph-Objekts sammeln:
    lKnoten.toFirst();
    lPos = 0;
    lNr = 1;
    if (MIT_XVERTEX) {
      while (lKnoten.hasAccess()) { 
        XVertex lTemp = (XVertex)lKnoten.getContent();
        lRet[lPos++] = "Knoten #" + (lNr++) + ": " + lTemp.getID() + " (" + lTemp.getX() + "," + lTemp.getY() + ")";
        lKnoten.next();
        //lNr++;
      } // end of while
    } else {
      while (lKnoten.hasAccess()) { 
        Vertex lTemp = lKnoten.getContent();
        lRet[lPos++] = "Knoten #" + (lNr++) + ": " + lTemp.getID();
        lKnoten.next();
        //lNr++;
      } // end of while
    } // end of if-else
    
    // Kantendaten des Graph-Objekts sammeln:
    lKanten.toFirst();
    lNr = 1;
    while (lKanten.hasAccess()) { 
      Edge lTemp = lKanten.getContent();
      Vertex[] lZweiKnoten = lTemp.getVertices(); 
      lRet[lPos++] = "Kante #" + (lNr++) + ": " + "K1=" + lZweiKnoten[0].getID() + " K2=" + lZweiKnoten[1].getID() + " Gewicht=" + lTemp.getWeight();
      lKanten.next();
      //lNr++;
    } // end of while
    return lRet;
  }            
  
  /**
   * @author Milan Paessler (Q1-IF-LK1-ZMK)
   */
  public List<Vertex> tiefenDurchlauf(String pStart) {
    this.meinGraph.setAllVertexMarks(false);
    
    Stack<Vertex> s = new Stack<Vertex>();    
    Vertex start = this.meinGraph.getVertex(pStart);
    start.setMark(true);
    s.push(start);
    
    List<Vertex> result = new List<Vertex>();
                  
    while (!s.isEmpty()) { 
      Vertex v1 = s.top();          
      s.pop();
      List<Vertex> neighbours = this.meinGraph.getNeighbours(v1);
      
      neighbours.toFirst(); 
      while (neighbours.hasAccess()) { 
        Vertex v2 = neighbours.getContent();
        neighbours.remove();
        if (!v2.isMarked()) {  
          v2.setMark(true);
          s.push(v2);     
          result.toLast();
          result.insert(v2);
        } // end of if
      } // end of for 
    } // end of while    
    
    return result;
  }
  
  /**
   * @author Milan Paessler (Q1-IF-LK1-ZMK)
   */
  public List<Vertex> breitenDurchlauf(String pStart) {
    this.meinGraph.setAllVertexMarks(false);
    
    Queue<Vertex> s = new Queue<Vertex>();    
    Vertex start = this.meinGraph.getVertex(pStart);
    start.setMark(true);
    s.enqueue(start);
    
    List<Vertex> result = new List<Vertex>();
                  
    while (!s.isEmpty()) { 
      Vertex v1 = s.front();          
      s.dequeue();
      List<Vertex> neighbours = this.meinGraph.getNeighbours(v1);
      
      neighbours.toFirst(); 
      while (neighbours.hasAccess()) { 
        Vertex v2 = neighbours.getContent();
        neighbours.remove();
        if (!v2.isMarked()) {  
          v2.setMark(true);
          s.enqueue(v2);     
          result.toLast();
          result.insert(v2);
        } // end of if
      } // end of for 
    } // end of while    
    
    return result;
  }
  
  // Ende Methoden
} // end of ProbeTest
